//------------------------------------------------------------------------------
// Прошивка для модуля-хаба (мост i2c-8x1-Wire)
// Код семейства - 0xA5
// Адрес на шине i2c - 0x15
// OldBean, 07.01.2019, 22.07.2019
//------------------------------------------------------------------------------
#include <avr/io.h>
#include <util/twi.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <string.h>
//------------------------------------------------------------------------------
#define F_CPU 16000000UL // Тактовая частота МК
#include <util/delay.h>
//------------------------------------------------------------------------------
#define ADDR 0x15        // Адрес устройства на шине I2C
#define FAMILY_CODE 0xA5 // Код семейства данного устройства
#define BUF_SIZE 72      // Размер буфера обмена данными
#define EE_SIZE 6        // Размер блока идентиф. и калибр. данных
//------------------------------------------------------------------------------
#define SCR_SIZE 9       // Длина блока данных 1-Wire устройств
#define CH_SCR_SIZE 16   // Область данных всех устройств
//------------------------------------------------------------------------------
void owReset();
void owWrite(uint8_t b);
void owRead(uint8_t *b);
static uint8_t bytes[8];               // Вспомогательный массив
static volatile uint8_t chMask = 0xff; // Маска каналов
static volatile uint8_t chSel = 0x0f;  // Номер выбранного канала (если один)
static volatile uint8_t scn = 8;       // Количество выделенных каналов  
static volatile uint8_t mask = 0xff;   // Рабочая маска
static volatile uint8_t c0750 = 0;     // Счетчик (время преобразования)
//------------------------------------------------------------------------------
// Команды:
// 0x33 Чтение кода семейства
// 0x4e Записать данные из буфера обмена в рабочую область (write scratch)
// 0xbe Считать данные из рабочей области в буфер обмена (read scratch)
// 0x48 Записать блок данных из рабочей области в EEPROM (copy scratch)
// 0xb8 Считать блок данных из EEPROM в рабочую область (recall EEPROM)
// 0x52 Запись данных из буфера обмена в рабочие регистры (для контроллеров)
// 0x44 Запуск преобразования (для датчиков)
//------------------------------------------------------------------------------
static uint8_t buf[BUF_SIZE];          // Буфер обмена
static uint8_t scratch[BUF_SIZE];      // Рабочая область RAM
static volatile uint8_t *ptr;          // Указатель на байт буфера обмена
static volatile uint8_t runFlag = 1;   // Рабочий режим, 0 - режим настройки
static volatile uint8_t nb = 0;        // Счетчик байтов
static volatile uint8_t prFlag = 0;    // Принят новый пакет байтов
static volatile uint8_t cmd = 0x00;    // Код команды
//------------------------------------------------------------------------------
ISR(TWI_vect) { // Обработчик событий модуля TWI
  switch(TW_STATUS) {     // Статус - в 5 старших битах регистра TWSR
    // Принимаем байты
    case TW_SR_SLA_ACK:   // Адрес для приема принят, ACK отправлен - 0x80
      ptr = buf;          // Настраиваемся на начало буфера обмена
      nb = 0;             // Сброс счетчика байтов
      break;
    case TW_SR_DATA_ACK:  // Очередной байт принят, ACK отправлен - 0x80
      *ptr++ = TWDR;
      nb++;
      break;
    case TW_SR_STOP:      // Состояние "стоп" на шине - 0xA0
      prFlag = 1;         // Весь пакет принят
      break;
    // Отправляем байты
    case TW_ST_SLA_ACK:   // Адрес для отправки принят, ACK отправлен - 0xA8
      ptr = buf;          // Настраиваемся на начало буфера обмена
      nb = 0;             // Сброс счетчика байтов
    case TW_ST_DATA_ACK:  // Очередной байт отправлен, ACK принят - 0xB8
      TWDR = *ptr++;
      nb++;
      break;
    case TW_ST_DATA_NACK: // Байт отправлен и получен NACK - 0xC0
    case TW_ST_LAST_DATA: // Отправлен последний байт, получен ACK - 0xC8
      break;
    // Ошибки старта или стопа - сброс TWI
    case TW_BUS_ERROR:
      TWCR = _BV(TWIE) | _BV(TWINT) | _BV(TWEA) | _BV(TWEN);
      break;
  }
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
}
//------------------------------------------------------------------------------
void owReset() { // Сброс и обнаружение подключенных к шинам 1-Wire датчиков
  PORTD = 0x00; DDRD = chMask; _delay_us(480); // Сброс всех линий
  PORTD = 0x00; DDRD = 0x00; _delay_us(60); // Отпустим линии
//  owdevs = ~PIND; // Это не имеет сойчас особого смысла - на всякий случай
  _delay_us(420); //ждем оставшееся время до 480мкс
}
//-----------------------------------------------------------------------------------------------------------
void owWrite(uint8_t b) { // Отсылаем один и тот же байт на все шины
  uint8_t mask = 0b00000001;
  for(uint8_t bit = 0; bit < 8; bit++) {
    cli();
    PORTD = 0x00; DDRD = chMask; _delay_us(3);  // Начало слота: все линии - в ноль
    if(b & mask) DDRD = 0x00; // Отпустим линии, если нужно передать единичку
    _delay_us(57);
    DDRD = 0x00; // Все равно отпускаем линии в конце слота
    sei();
    _delay_us(5);
    mask <<= 1;
  }
}
//-----------------------------------------------------------------------------
void owRead(uint8_t *b) { // Чтение байтов со всех линий
  uint8_t wb[8]; // Вспомогательный массив, который заносятся читаемые биты.
                 // Каждый байт wb содержит соответствующие биты всех датчиков.
                 // Например, wb[3] содержит 3-и биты всех датчиков.
  for(uint8_t bit = 0; bit < 8; bit++) { // Пробегаем 
    cli();
    PORTD = 0x00; DDRD = chMask; _delay_us(3); // Начало слота: линии - в ноль
    DDRD = 0x00; _delay_us(12); // Отпустим все линии и пауза
    wb[bit] = PIND; // Читаем со всех линий биты с номером bit
    _delay_us(45); // Пауза до конца слота
    sei();
  }
  _delay_us(5);
  // Преобразуем полученную информацию (массив wb) к нормальной форме, чтобы
  // каждый байт содержал все биты, полученные от одного датчика. По сути,
  // нам нужно просто транспонировать матрицу 8x8, представленную массивом wb.
  uint8_t bmask = 0b00000001;
  for(uint8_t bit = 0; bit < 8; bit++) {
    uint8_t nmask = 0b00000001;
    for(uint8_t sn = 0; sn < 8; sn++) {
      if(wb[bit] & nmask) b[sn] |= bmask; // Установим бит
      else b[sn] &= ~bmask; // Сбросим бит
      nmask <<= 1;
    }
    bmask <<= 1;
  }
}
//------------------------------------------------------------------------------
ISR(TIMER0_OVF_vect) { // Обработчик по переполнению таймера 0
  if(++c0750 == 47) { // Преобразование закончено, данные готовы - читаем их
    TIMSK0 &= ~_BV(TOIE0); // Запрещаем прерывания по переполнению
    owReset();
    owWrite(0xcc); // Skip ROM
    owWrite(0xbe); // Будем читать данные
    owRead(bytes);
    for(uint8_t i = 0; i < 8; i++)
      scratch[2*i] = bytes[i];
    owRead(bytes);
    for(uint8_t i = 0; i < 8; i++)
      scratch[1 + 2*i] = bytes[i];
    owReset();
    memcpy(buf, scratch, CH_SCR_SIZE);
    TWCR |= _BV(TWEN); // Активируем TWI
  }
}
//------------------------------------------------------------------------------
int main(void) {
  // Инициализация TWI-модуля --------------------------------------------------
  TWAR = ADDR << 1;   // Адрес устройства - в старших 7 битах
  TWCR = _BV(TWIE);   // Разрешаем прерывания от TWI
  TWCR |= _BV(TWEA);  // Разрешаем подачу импульса ACK
  TWCR |= _BV(TWINT); // Очистим флаг TWINT (очищается при установке в 1)
  TWCR |= _BV(TWEN);  // Разрешаем работу TWI и активирует TWI-интерфейс
  // Инициализация таймера 0 для долговременного тактирования ------------------
  // Счетчик работает в обычном режиме (Normal mode)
  TCCR0B = _BV(CS02) | _BV(CS00); // Предделитель на 1024
  TIMSK0 &= ~_BV(TOIE0); // Запрещаем прерывания по переполнению

  sei(); // Разрешаем глобальные прерывания
  uint8_t index; // Рабочая переменная
  
  ptr = buf;
  while(1) { // Бесконечный цикл
    if(prFlag) { // По шине i2c принят новый пакет - отрабатываем его
      TWCR &= ~_BV(TWEN); // Приостановим работу TWI
      prFlag = 0; // Сбросим флаг получения нового пакета
      cmd = buf[0]; // В первом принятом байте - код команды
      switch(cmd) { // Анализируем и выполняем команды
        case 0x33: // Чтение ROM - в первом байте - код семейства
          buf[0] = FAMILY_CODE; // Код семейства данного устройства
          owReset(); // Получим коды семейств устройств, подключенных к хабу
          owWrite(0x33);
          owRead(bytes);
          owReset();
          memcpy(&buf[1], bytes, 8);
          TWCR |= _BV(TWEN); // Активируем TWI
          break;
        case 0x4e: // write scratch - записать данные в рабочую область устр-ва
          chSel = buf[1]; // Номер канала устройства - во втором байте команды
          if(chSel > 7) break;
          chMask = _BV(chSel); // Работаем только с выбранным каналом
          owReset();
          owWrite(0xcc);
          owWrite(0x4e);
          for(uint8_t i = 0; i < EE_SIZE; i++)
            owWrite(buf[i + 2]);
          chMask = 0xff; // Восстановим маску каналов
          TWCR |= _BV(TWEN); // Активируем TWI
          break;
        case 0xbe: // read scratch - считать данные из рабочих областей устр-в
          owReset();
          owWrite(0xcc);
          owWrite(0xbe);
          for(uint8_t i = 0; i < SCR_SIZE; i++) {
            owRead(bytes);
            index = i;
            for(uint8_t j = 0; j < 8; j++) {
              buf[index] = bytes[j];
              index += SCR_SIZE;
            }
          }
          TWCR |= _BV(TWEN); // Активируем TWI
          break;
        case 0x48: // copy scratch - записать данные на устройстве в EEPROM
          chSel = buf[1]; // Номер канала устройства - во втором байте команды
          if(chSel > 7) break;
          chMask = _BV(chSel); // Работаем только с выбранным каналом
          owReset();
          owWrite(0xcc);
          owWrite(0x48);
          chMask = 0xff; // Восстановим маску каналов
          TWCR |= _BV(TWEN); // Активируем TWI
          break;
        case 0xb8: // recall EEPROM - считать из EEPROM на устройствах
          owReset();
          owWrite(0xcc);
          owWrite(0xb8);
          TWCR |= _BV(TWEN); // Активируем TWI
          break;
        case 0x52: // Запись данных в регистры (для контроллеров)
          TWCR |= _BV(TWEN); // Активируем TWI
          break;
        case 0x44: // Запуск преобразования
          owReset();
          owWrite(0xcc);
          owWrite(0x44);
          c0750 = 0; // Сбрасываем счетчик для задания времени преобразования
          TIMSK0 |= _BV(TOIE0); // Разрешим прерывания по переполнению таймера 0
          // Пока преобразование не закончится TWI не активен (см. обработчик
          // прерываний по переполнению таймера 0 (ISR(TIMER0_OVF_vect))
          break;
        default:
          TWCR |= _BV(TWEN); // Активируем TWI
      }
      prFlag = 0; // Сбросим флаг получения нового пакета
    }
  }
  return 0;
}
//------------------------------------------------------------------------------
// История и список изменений
// 1. 29.11.2018 - Вчерне реализован весь функционал для этого типа устройств.
// 2. 07.01.2019 - Отладка блока калибровки и измерений
// 3. 22.07.2019 - Изменена логика запуска преобразования датчиков (непрерывный
//                 запуск преобразования удален)
//               - Добавлена логика для TWI "готов/не готов"
//------------------------------------------------------------------------------
